﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace mp3_14_ctrl {

    public partial class Form1: Form {
#if DEBUG
        private const String LOCAL_DIR = @"C:\\UserData\Source\Ele\MP3プレーヤ14(RaspberryPIZeroW+TWELITE)\prg\";
#else
        private const String LOCAL_DIR = "";
#endif

        private const String DIRSTRUCTUREFILE14 = LOCAL_DIR + "DIRLST14.DAT";  // DIR構成ファイル(Version=14)
        private const int RECLEN_DIRLIST = 768;                     // レコード長(0x500)
        private const int DL_FULLDIR_OFFSET = 48;					// ディレクトリフルパス
        private const int DL_LENFULLDIR_OFFSET = 561;               // ディレクトリフルパスバイト数
        private const int DL_RECNO_SUBDIR_OFFSET = 691;				// サブディレクトリの先頭レコード番号
        private const int DL_NUM_SUBDIR_OFFSET = 693;   			// サブディレクトリ数
        private const int DL_RECNO_PARENTDIR_OFFSET = 694;  		// 親ディレクトリのレコード番号
        private const int DL_RECNO_GRANDPARENTDIR_OFFSET = 696;		// 親々ディレクトリのレコード番号
        private const int DL_NUM_SNDFILE_OFFSET = 698;				// MP3ファイル数
        private const int DL_RECNO_SNDFILE_OFFSET = 699;			// MP3ファイルリストの最初の曲のレコード番号

        private const string SNDFILELISTFILE = LOCAL_DIR + "SNDLST14.DAT";     // MP3ファイルリスト(Version=14)
        private const int RECLEN_MP3LIST = 512;                     // イメージ部分を含むレコード長
        private const int ML_FILENAME_OFFSET = 128;                 // 

        // データファイル取得バッチとスクリプトファイル(本PGの設定定義ファイルを兼ねる)
        private const String GET_DATAFILE_BAT = LOCAL_DIR + "get_datafile_from_host.bat";
        private const String GET_DATAFILE_SCR = LOCAL_DIR + "get_datafile_from_host.txt";

        // ウインドウズ版コントローラとの通信コマンドコード(mp3_14_menu.cと対応)
        private const Byte WINCTRL_CMD_PLAYDIR = 0x01;
        private const Byte WINCTRL_CMD_PLAY = 0x02;
        private const Byte WINCTRL_CMD_PLAYRANDOM = 0x03;
        private const Byte WINCTRL_CMD_VOLUP = 0x04;
        private const Byte WINCTRL_CMD_VOLDOWN = 0x05;
        private const Byte WINCTRL_CMD_PLAYSTOP = 0x06;
        private const Byte WINCTRL_CMD_PLAYSKIP = 0x07;
        private const Byte WINCTRL_CMD_ACTIVE = 0x08;
        private const Byte WINCTRL_CMD_NOWPLAYING = 0x09;
        private const Byte WINCTRL_CMD_WAITING = 0x0A;
        private const Byte WINCTRL_CMD_NEXT = 0x0B;
        private const Byte WINCTRL_CMD_PREV = 0x0C;
        private const Byte WINCTRL_CMD_SHUTDOWN = 0x0D;
        private const Byte WINCTRL_CMD_FILESIZE = 0x0E;
        private const Byte WINCTRL_CMD_EXECMKFL = 0x0F;
        private const Byte WINCTRL_CMD_DATATIMESTAMP = 0x10;
        private const Byte WINCTRL_CMD_HALT = 0x11;

        private const int VOLUME_STEP = 4;                  // 音量の増減単位(mp3_14_menu.cのVOLUME_STEPと同じ値)
        private const int WINCTRL_CMDLENGTH = 16;           // ソケットを通じて送受信する１単位のデータサイズ(BYTE)
        private const Byte NOWPLAYING_STARTEDNOW = 0x01;    // WINCTRL_CMD_NOWPLAYINGのフラグパラメータ：曲の再生開始
        private static Color COLOR_VOLUME_UNMATCH = Color.LightCoral;  // 音量不一致の色
        private const int MIN_VOLUME = -175;                // MP3デコーダ madplay の制限
        private const int MAX_VOLUME = +18;                 // MP3デコーダ madplay の制限
        private const int WHEEL_UPDOWN_TIME = 30000;        // ホイールによる時間移動単位 ms

        private static UInt16 DirListDetailRecNum;	// DIR構成ファイルの明細レコード総数
        private static DIRLIST[] DirList;           // DIR構成ファイル
        private static bool DirListLoaded;          // DIR構成ファイルをTreeViewに設定済みか
        private static FileStream SndListReader = null;    // MP3ファイルリスト
        private static TcpClient connectToHost = null;      // mp3_14_servとの通信
        private static NetworkStream streamToHost = null;   // connectToHostのストリーム
        private static bool NowPlaying = false;     // 再生中のときtrue
        private static bool StartUp = false;        // 
        private static UInt32 PassedTime;           // プログラム起動後経過時間（ms）
        private static UInt32 CurTrackPlayingTime;  // 現在のトラックの再生経過時間ms
        private static UInt32 CurTrackTotalTime;    // 現在のトラックの総時間ms
        private static UInt32 TxtVolumeFocusedTime; // 
        private static UInt16 CurPlayingDirno;      // 再生中のDirNo
        private static UInt16 CurPlayingTrackno;    // 再生中のトラックNo
        private static int CurVolume;               // 再生中の音量
        private static int CmdActiveCount;          // WINCTRL_CMD_ACTIVEの送信頻度制御
        private static UInt32 Mp3FileSize;          // MP3ファイルのバイトサイズ
        private static bool DataTimestampMatched;   // データファイルのタイムスタンプ一致した
        private static bool WaitCmdDataTimestampFlag;   // WINCTRL_CMD_DATATIMESTAMPの結果待ちフラグ
        private static String DataTimeStampHost;    // HOST側データファイルのタイムスタンプ
        private static String DataTimeStampLocal;   // LOCAL側データファイルのタイムスタンプ

        public Form1()
        {
            InitializeComponent();
            txtVolume.MouseWheel += new System.Windows.Forms.MouseEventHandler(this.txtVolume_MouseWheel);
            txtPlayTime.MouseWheel += new System.Windows.Forms.MouseEventHandler(this.txtPlayTime_MouseWheel);
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            SetControlsNoConnectToHost();
            txtVolume.Text = CurVolume.ToString();

            if (ConnectToHost() == true) {
                StartCompareDataTimestamp();
            }

        }

        /// <summary>
        /// ホストに接続
        /// </summary>
        /// <returns></returns>
        private bool ConnectToHost()
        {
            try {
                const String KEY_SERVER = "MP3_14_SERVER=";
                const String KEY_PORT = "MP3_14_PORT=";

                String server = "";
                String port = "";

                using (StreamReader sr = new StreamReader(GET_DATAFILE_SCR)) {
                    string line;
                    int idx;

                    while ((line = sr.ReadLine()) != null) {
                        if (line.Length > 0 && line[0] == '#') {
                            Console.WriteLine(line);
                            idx = line.IndexOf(KEY_SERVER);
                            if (idx > 0) {
                                server = line.Substring(idx + KEY_SERVER.Length);
                            } else {
                                idx = line.IndexOf(KEY_PORT);
                                if (idx > 0) {
                                    port = line.Substring(idx + KEY_PORT.Length);
                                    if (server != "") break;
                                }
                            }
                        }
                    }
                }
                connectToHost = new TcpClient(server, Int16.Parse(port));
//                connectToHost = new TcpClient("192.168.11.9", 5000);
                streamToHost = connectToHost.GetStream();

                return true;
            } catch (SocketException ex) {
                MessageBox.Show("MP3_14ホストにアクセスできません。" + ex.Message);
                return false;
            }

        }

        /// <summary>
        /// DIR構成ファイルを読み込む 
        /// </summary>
        private void LoadDirStructFile()
        {
            Decoder d = Encoding.UTF8.GetDecoder();
            byte[] RecDirList = new byte[RECLEN_DIRLIST];
            int numBytesToRead = (int)RecDirList.Length;
            int idxout;
            DIRLIST dir;
            FileStream reader = null;
            int dirno;

            // ディレクトリ構成ファイルオープン
            try {
                reader = new FileStream(DIRSTRUCTUREFILE14, FileMode.Open, FileAccess.Read);
            } catch (FileNotFoundException ex) {
                MessageBox.Show("ディレクトリ構成ファイルが存在しません");
                return;
            }
            // Skip HEADER info
            if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) != RECLEN_DIRLIST) {
                MessageBox.Show("ディレクトリ構成ファイルが不正です");
                reader.Close();
                return;
            }
            dirno = 0;
            DirListDetailRecNum = (UInt16)(RecDirList[0] | ((UInt16)RecDirList[1] << 8));       // ディレクトリ総数
            DirList = new DIRLIST[DirListDetailRecNum];
            idxout = 0;

            while (true) {
                if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) != RECLEN_DIRLIST) {
                    break;
                }
                dirno++;

                dir = new DIRLIST();

                int bytes;
                // 
                bytes = RecDirList[DL_LENFULLDIR_OFFSET] | ((UInt16)RecDirList[DL_LENFULLDIR_OFFSET + 1] << 8);
                int cn = d.GetCharCount(RecDirList, DL_FULLDIR_OFFSET, bytes);
                Char[] chars = new char[cn];
                d.GetChars(RecDirList, DL_FULLDIR_OFFSET, bytes, chars, 0);
                dir.fulldir = new String(chars);        // ディレクトリフルパス
                //
                int lastidx = dir.fulldir.LastIndexOf('/');
                if (lastidx > 0) {
                    dir.dirname = dir.fulldir.Substring(lastidx + 1);
                } else {
                    dir.dirname = "???";
                }
                //
                dir.recno_subdir_start = (UInt16)(RecDirList[DL_RECNO_SUBDIR_OFFSET] + ((UInt16)RecDirList[DL_RECNO_SUBDIR_OFFSET + 1] << 8) - 1);  // ヘッダレコードの1行分を引く
                dir.num_subdir = (UInt16)(RecDirList[DL_NUM_SUBDIR_OFFSET]);
                dir.recno_parentdir = (UInt16)(RecDirList[DL_RECNO_PARENTDIR_OFFSET] + ((UInt16)RecDirList[DL_RECNO_PARENTDIR_OFFSET + 1] << 8) - 1);
                dir.recno_grandparentdir = (UInt16)(RecDirList[DL_RECNO_GRANDPARENTDIR_OFFSET] + ((UInt16)RecDirList[DL_RECNO_GRANDPARENTDIR_OFFSET + 1] << 8) - 1);
                dir.num_sndfile = (UInt16)(RecDirList[DL_NUM_SNDFILE_OFFSET]);
                dir.recno_sndfile = (UInt16)(RecDirList[DL_RECNO_SNDFILE_OFFSET] + ((UInt16)RecDirList[DL_RECNO_SNDFILE_OFFSET + 1] << 8));     // MP3ファイルリストファイルはヘッダレコードなし
                dir.dirno = (UInt16)dirno;
                DirList[idxout] = dir;

                System.Console.WriteLine(idxout + " バイト数=" + bytes + " fulldir=" + dir.fulldir + " dirname=" + dir.dirname + " recno_subdir_start=" + dir.recno_subdir_start + " num_subdir=" + dir.num_subdir);
                idxout++;
            }
            reader.Close();
        }

        /// <summary>
        /// ディレクトリ構造をTreeViewに表示 
        /// </summary>
        private void SetDirTreeView()
        {
            int i;
            int dirs;
            int icono;      // ICON No.
            DIRLIST dir = DirList[0];
            ImageList iconImageList = new ImageList();
            iconImageList.Images.Add(Properties.Resources.cd1);
            iconImageList.Images.Add(Properties.Resources.person1);
            iconImageList.Images.Add(Properties.Resources.folder);
            treeViewDir.ImageList = iconImageList;

            dirs = dir.num_subdir;
            for (i = 0; i < dirs; i++) {
                DIRLIST subdir = DirList[dir.recno_subdir_start + i];
                if (subdir.num_sndfile > 0) {
                    icono = 0;
                } else {
                    icono = 2;
                }
                TreeNode dirNode = new TreeNode(subdir.dirname, icono, icono);
                dirNode.Tag = subdir;
                treeViewDir.Nodes.Add(dirNode);
                subdir.treenode = dirNode;

                AddDirTreeViewNode(dirNode, subdir);
            }
        }

        /// <summary>
        /// ディレクトリ構造にノードを再帰的に追加 
        /// </summary>
        /// <param name="DirNode"></param>
        /// <param name="dir"></param>
        private void AddDirTreeViewNode(TreeNode DirNode, DIRLIST dir)
        {
            int dirs = dir.num_subdir;
            int i;
            int icono;

            for (i = 0; i < dirs; i++) {
                DIRLIST subdir = DirList[dir.recno_subdir_start + i];
                if (subdir.num_sndfile > 0) {
                    icono = 0;
                } else {
                    icono = 1;
                }
                TreeNode dirNode = new TreeNode(subdir.dirname, icono, icono);
                dirNode.Tag = subdir;
                DirNode.Nodes.Add(dirNode);
                subdir.treenode = dirNode;

                AddDirTreeViewNode(dirNode, subdir);
            }
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            try {
                SndListReader.Close();
                if (connectToHost != null) {
                    connectToHost.Close();
                }
            } catch (Exception err) {

            }
        }

        /// <summary>
        /// ディレクトリ選択時
        /// 手で選択したときだけでなく、プログラムから選択したときも発生する
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void treeViewDir_AfterSelect(object sender, TreeViewEventArgs e)
        {
            TreeNode dirNode = e.Node;
            bool btnPlayEnabled = false;

            if (dirNode != null) {
                DIRLIST dir = (DIRLIST)dirNode.Tag;
                if (dir != null) {
                    listViewMp3.Items.Clear();
                    // アルバムの曲目を表示
                    Decoder d = Encoding.UTF8.GetDecoder();
                    byte[] RecMp3List = new byte[RECLEN_MP3LIST];
                    SndListReader.Seek(dir.recno_sndfile * RECLEN_MP3LIST, SeekOrigin.Begin);
                    for (int i = 0; i < dir.num_sndfile; i++) {
                        if (SndListReader.Read(RecMp3List, 0, RECLEN_MP3LIST) != RECLEN_MP3LIST) {
                            break;
                        }
                        int cn = d.GetCharCount(RecMp3List, ML_FILENAME_OFFSET, 193);
                        Char[] chars = new char[cn];
                        d.GetChars(RecMp3List, ML_FILENAME_OFFSET, ML_FILENAME_OFFSET, chars, 0);
                        ListViewItem item = new ListViewItem(new String(chars));
                        listViewMp3.Items.Add(item);
                    }
                    if (dir.num_sndfile > 0) {
                        listViewMp3.Focus();
                        listViewMp3.Items[0].Selected = true;
                        btnPlayEnabled = true;
                    } else {
                    }
                } else {
                    listViewMp3.Items.Clear();
                }
            } else {
                listViewMp3.Items.Clear();
            }

            if (NowPlaying == false) {
                btnPlay.Enabled = btnPlayEnabled;
            }
        }


        /// <summary>
        /// PLAYボタン（再生開始）
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnPlay_Click(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            try {
                if (NowPlaying == false && DirListLoaded == true) {
                    if (chkRandom.Checked == false) {
                        DIRLIST dir = (DIRLIST)(treeViewDir.SelectedNode.Tag);
                        Byte trackno = 0;
                        ushort start_time = 0;
                        if (listViewMp3.SelectedItems.Count > 0) {
                            trackno = (Byte)(listViewMp3.SelectedItems[0].Index);
                        } else {
                            return;
                        }

                        if (dir.dirno == CurPlayingDirno && trackno == CurPlayingTrackno) {
                            start_time = (ushort)(CurTrackPlayingTime / 1000);
                            SetPlayTime(start_time);
                        } else {
                            CurTrackPlayingTime = 0;
                            SetPlayTime(0);
                        }
                        buf[0] = WINCTRL_CMD_PLAYDIR;
                        buf[1] = (byte)(dir.dirno & 0x00FF);
                        buf[2] = (byte)((dir.dirno >> 8) & 0x00FF);
                        buf[3] = trackno;
                        buf[4] = (byte)(start_time & 0x00FF);
                        buf[5] = (byte)((start_time >> 8) & 0x00FF);
                        streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                        txtVolume.BackColor = Color.White;

                        btnSkip.Enabled = true;
                    } else {
                        ushort start_time = (ushort)(CurTrackPlayingTime / 1000);
                        SetPlayTime(start_time);
                        buf[0] = WINCTRL_CMD_PLAYRANDOM;
                        buf[1] = (byte)(start_time & 0x00FF);
                        buf[2] = (byte)((start_time >> 8) & 0x00FF);
                        streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                        btnSkip.Enabled = true;
                        //CurPlayingTime = 0;
                        //SetPlayTime(0);
                    }

                    CurTrackTotalTime = 0;
                    NowPlaying = true;
                    btnPlay.Enabled = false;
                    btnPrev.Enabled = true;
                    btnNext.Enabled = true;
                    btnStop.Enabled = true;
                    chkRandom.Enabled = false;
                    listViewMp3.Focus();
                    // ↑この時点ではtxtVolume.Enabled = trueにしない。
                }
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }

        /*
         * タイマーイベント
         */
        private void timer1_Tick(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];
            int val;

            try {
                PassedTime += (UInt32)timer1.Interval;

                if (NowPlaying == true) {
                    CurTrackPlayingTime += (UInt32)timer1.Interval;
                    // 音量テキストボックスのフォーカスを外す
                    if (TxtVolumeFocusedTime > 0 && (PassedTime - TxtVolumeFocusedTime) > 3000) {   // 音量変更後3秒経過
                        TxtVolumeFocusedTime = 0;
                        listViewMp3.Focus();
                    }
                }

                // ホストからの受信
                while (streamToHost != null && streamToHost.DataAvailable == true) {
                    val = Read_Mp3_14_host(buf);
                    if (val > 0) {
                        System.Console.WriteLine("RECV from HOST cmd=" + buf[0] + ", bytes=" + val);
                    }
                    if (val == WINCTRL_CMDLENGTH && buf[0] == WINCTRL_CMD_NOWPLAYING) {
                        UInt16 dirno;
                        UInt16 trackno;
                        UInt16 playtime;
                        UInt32 volume;
                        UInt32 totaltime;
                        Byte flag;

                        System.Console.WriteLine("WINCTRL_CMD_NOWPLAYING(" + WINCTRL_CMD_NOWPLAYING + ") received " + buf[1].ToString("X") + "," + buf[2].ToString("X") + ",");

                        dirno = (UInt16)(buf[1] | ((UInt16)buf[2]) << 8);
                        trackno = buf[3];
                        playtime = (UInt16)(buf[4] | ((UInt16)buf[5]) << 8);
                        volume = (UInt32)(buf[6] | ((UInt32)buf[7]) << 8 | ((UInt32)buf[8]) << 16 | ((UInt32)buf[9]) << 24);
                        flag = buf[10];
                        totaltime = (UInt32)(buf[11] | ((UInt32)buf[12]) << 8);
                        //System.Console.WriteLine("dirno=" + dirno + " trackno=" + trackno + " volume=" + (int)volume + " CurVolume=" + CurVolume + " Play=" + playtime + " Total=" + totaltime);
                        NowPlaying = true;
                        SetControlsNowPlaying();

                        if (DirListLoaded) {
                            if (CurPlayingDirno != dirno || CurPlayingTrackno != trackno
                                || treeViewDir.SelectedNode != DirList[CurPlayingDirno - 1].treenode) {

                                CurPlayingDirno = dirno;
                                CurPlayingTrackno = trackno;
                                System.Console.WriteLine(" CurPlayingDirno=" + CurPlayingDirno + " CurPlayingTrackno=" + CurPlayingTrackno);
                                // 現在再生中のディレクトリ番号に従ってtreeViewDirの再選択
                                if (CurPlayingDirno >= 1 && CurPlayingDirno <= DirListDetailRecNum) {
                                    treeViewDir.SelectedNode = DirList[CurPlayingDirno - 1].treenode;
                                    System.Console.WriteLine(DirList[CurPlayingDirno - 1].fulldir);
                                }
                                // ファイルサイズ要求
                                //buf[0] = WINCTRL_CMD_FILESIZE;
                                //streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                                //System.Console.WriteLine("WINCTRL_CMD_FILESIZE sent ");
                                Mp3FileSize = 0;
                                pgbPlayTime.Value = 0;
                            }
                        }
                        // スライドバー位置
                        if (totaltime != 0) {
                            int value = (int)(((float)playtime / (float)totaltime) * 100);
                            if (value > 100) value = 100;
                            pgbPlayTime.Value = value;
                        }
                        // 再生経過時間
                        if (flag == NOWPLAYING_STARTEDNOW || (playtime * 1000) >= CurTrackPlayingTime) {
                            CurTrackPlayingTime = (UInt32)(playtime * 1000);
                            CurTrackTotalTime = (UInt32)(totaltime * 1000);
                        }
                        // 音量
                        if (flag == NOWPLAYING_STARTEDNOW || StartUp == false) {
                            CurVolume = (int)volume;
                            txtVolume.Text = CurVolume.ToString();
                            txtVolume.BackColor = Color.White;
                            txtVolume.Enabled = true;
                            StartUp = true;
                        } else if (CurVolume != (int)volume) {       // mp3_14_menuが保持しているボリュームと本PGが保持してるボリュームが異なる
                            txtVolume.BackColor = COLOR_VOLUME_UNMATCH;
                        } else {
                            txtVolume.BackColor = Color.White;
                        }
                        if (trackno < listViewMp3.Items.Count) {
                            listViewMp3.Items[trackno].Selected = true;
                        }
                    } else if (buf[0] == WINCTRL_CMD_FILESIZE) {
                        // ファイルサイズ取得
                        Mp3FileSize = (UInt32)(buf[1] | ((UInt32)buf[2]) << 8 | ((UInt32)buf[3]) << 16 | ((UInt32)buf[4]) << 24);
                        System.Console.WriteLine("WINCTRL_CMD_FILESIZE recieved " + Mp3FileSize);
                    } else if (buf[0] == WINCTRL_CMD_WAITING) {
                        // ホストは再生していない
                        System.Console.WriteLine("WINCTRL_CMD_WAITING recieved");
                        NowPlaying = false;
                        SetControlsCmdWaiting();
                    } else if (buf[0] == WINCTRL_CMD_SHUTDOWN) {
                        // ホストが停止したことの通知を受ける
                        System.Console.WriteLine("WINCTRL_CMD_SHUTDOWN recieved");
                        NowPlaying = false;
                        SetControlsNoConnectToHost();
                        if (connectToHost != null) {
                            streamToHost.Close();
                            streamToHost = null;
                            connectToHost.Close();
                            connectToHost = null;
                        }
                        //
                        MessageBox.Show("ホストが停止しました。");
                        this.Close();       // アプリケーション自体停止
                    } else if (buf[0] == WINCTRL_CMD_EXECMKFL) {
                        // ホストでのmp3_14_mkfl実行が終了
                        try {
                            SndListReader.Close();
                        } catch (Exception err) {
                        }
                        try {
                            System.Diagnostics.Process p = System.Diagnostics.Process.Start(GET_DATAFILE_BAT);
                            p.WaitForExit();
                        } catch (Exception err) {
                            MessageBox.Show("データファイルの作成でエラー：" + err.Message);
                        }
                        MessageBox.Show("本プログラムを再起動してください。");

                    } else if (buf[0] == WINCTRL_CMD_DATATIMESTAMP && WaitCmdDataTimestampFlag == true) {
                        System.Console.WriteLine("WINCTRL_CMD_DATATIMESTAMP recieved");
                        DataTimeStampHost = Encoding.UTF8.GetString(buf, 1, 14);
                        System.Console.WriteLine("WINCTRL_CMD_DATATIMESTAMP recieved " + DataTimeStampHost);
                        if (DataTimeStampHost.Equals(DataTimeStampLocal) == true) {
                            DataTimestampMatched = true;
                            LoadDirStructFile();
                            SetDirTreeView();
                            DirListLoaded = true;
                            try {
                                SndListReader = new FileStream(SNDFILELISTFILE, FileMode.Open, FileAccess.Read);
                                // 接続できた
                                SetControlsConnectToHost();
                            } catch (FileNotFoundException ex) {
                                MessageBox.Show("MP3ファイルリストファイルが存在しません。");
                            }
                        } else {
                            DataTimestampMatched = false;
                            MessageBox.Show("データファイルのタイムスタンプが一致しません");
                        }
                        WaitCmdDataTimestampFlag = false;
                    }
                }

                if (NowPlaying == true) {
                    SetPlayTime((UInt16)(CurTrackPlayingTime / 1000));
                }

                // ホストへ送信
                if (connectToHost != null) {
                    CmdActiveCount++;
                    if (CmdActiveCount == 5) {      // 送信頻度調整
                        CmdActiveCount = 0;
                        System.Console.WriteLine("WINCTRL_CMD_ACTIVE(" + WINCTRL_CMD_ACTIVE + ") sent");
                        buf[0] = WINCTRL_CMD_ACTIVE;
                        streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                    }
                }
            } catch (Exception ex) {
                System.Console.WriteLine(ex.Message);
            }
        }


        /// <summary>
        /// mp3_14_servからのデータ取得
        /// </summary>
        /// <param name="buf"></param>
        /// <returns></returns>
        private int Read_Mp3_14_host(byte[] buf)
        {
            int total_size = 0;
            int val;

            while (total_size != WINCTRL_CMDLENGTH) {
                val = streamToHost.Read(buf, 0, WINCTRL_CMDLENGTH - total_size);
                total_size += val;
            }
            return total_size;
        }


        /// <summary>
        /// 再生経過時間を表示
        /// </summary>
        /// <param name="playtime">秒</param>
        private void SetPlayTime(UInt16 playtime)
        {
            int sec = (playtime % 60);      // 秒
            int min = (playtime - sec) / 60;

            txtPlayTime.Text = min.ToString("00") + ":" + sec.ToString("00");
        }

        private void txtPlayTime_MouseWheel(object sender, MouseEventArgs e)
        {
            try {
                if (NowPlaying == false) {
                    if (e.Delta > 0) {      // 進める
                        if ((CurTrackPlayingTime + WHEEL_UPDOWN_TIME) < CurTrackTotalTime) {
                            CurTrackPlayingTime += WHEEL_UPDOWN_TIME;
                            SetPlayTime((UInt16)(CurTrackPlayingTime / 1000));
                        }
                    } else if (e.Delta < 0) {   //  戻す
                        if (CurTrackPlayingTime > WHEEL_UPDOWN_TIME) {
                            CurTrackPlayingTime -= WHEEL_UPDOWN_TIME;
                        } else {
                            CurTrackPlayingTime = 0;
                        }
                        SetPlayTime((UInt16)(CurTrackPlayingTime / 1000));
                    } else {
                        return;
                    }
                }
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }


        /// <summary>
        /// 再生を停止。同じ位置から再開することがありえるのでCurPlayingTimeをリセットしないこと
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnStop_Click(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            try {
                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_PLAYSTOP;
                    streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);

                    //NowPlaying = false;
                    //SetControlsCmdWaiting();
                }
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }

        /// <summary>
        /// ランダム再生チェックボックス　オンオフ切り替え
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void chkRandom_CheckedChanged(object sender, EventArgs e)
        {
            if (chkRandom.Checked == true) {
                btnPlay.Enabled = true;
                //CurPlayingTime = 0;
                //SetPlayTime(0);
            } else {    // ランダム再生オフ
                // MP3ファイルを持つディレクトリを選択しているならPLAYボタンを使用可能にする
                if (treeViewDir.SelectedNode != null) {
                    DIRLIST dir = (DIRLIST)(treeViewDir.SelectedNode.Tag);
                    if (dir != null) {
                        if (dir.num_sndfile > 0) {
                            btnPlay.Enabled = true;
                        } else {
                            btnPlay.Enabled = false;
                        }
                    } else {
                        btnPlay.Enabled = false;
                    }
                } else {
                    btnPlay.Enabled = false;
                }
            }
            CurTrackTotalTime = 0;
            CurTrackPlayingTime = 0;
            SetPlayTime(0);
        }

        /// <summary>
        /// スキップボタン　ランダム再生時に次のディレクトリの再生へ進む
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnSkip_Click(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            try {
                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_PLAYSKIP;
                    streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                }
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }


        /// <summary>
        /// コントローラからホスト接続、またはコントローラからホスト停止
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnStartStop_Click(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            try {
                if (connectToHost == null) {            // ホストに接続してなければ
                    if (ConnectToHost() == true) {      // 接続して
                        StartCompareDataTimestamp();    // データファイルが最新かを確認
                    }

                } else {
#if DEBUG
                    buf[0] = WINCTRL_CMD_HALT;
#else
                    buf[0] = WINCTRL_CMD_SHUTDOWN;
#endif
                    streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                }
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }

        /// <summary>
        /// 前の曲
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnPrev_Click(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            try {
                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_PREV;
                    streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                }
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }


        /// <summary>
        /// 次の曲
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnNext_Click(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            try {
                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_NEXT;
                    streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                }
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }

        /// <summary>
        /// 音量
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void txtVolume_Enter(object sender, EventArgs e)
        {
            TxtVolumeFocusedTime = PassedTime;
        }

        /// <summary>
        /// 音量をマウスホイールで変更
        /// -175 - +18
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void txtVolume_MouseWheel(object sender, MouseEventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];
            try {
                System.Console.WriteLine("delta=" + e.Delta + " lines-" + SystemInformation.MouseWheelScrollLines);     // スクロール量（方向）の表示  
                if (e.Delta > 0) {
                    buf[0] = WINCTRL_CMD_VOLUP;
                    if ((Control.ModifierKeys & Keys.Control) == Keys.Control) {
                        buf[1] = 1;
                    } else {
                        buf[1] = VOLUME_STEP;
                    }
                    streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                    CurVolume += buf[1];
                    //CurVolume += VOLUME_STEP;
                    if (CurVolume > MAX_VOLUME) CurVolume = MAX_VOLUME;
                } else if (e.Delta < 0) {
                    buf[0] = WINCTRL_CMD_VOLDOWN;
                    if ((Control.ModifierKeys & Keys.Control) == Keys.Control) {
                        buf[1] = 1;
                    } else {
                        buf[1] = VOLUME_STEP;
                    }
                    streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                    CurVolume -= buf[1];
                    //CurVolume -= VOLUME_STEP;
                    if (CurVolume < MIN_VOLUME) CurVolume = MIN_VOLUME;
                } else {
                    return;
                }
                txtVolume.Text = CurVolume.ToString();
                txtVolume.BackColor = COLOR_VOLUME_UNMATCH;
                TxtVolumeFocusedTime = PassedTime;
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }
        }

        /// <summary>
        /// 音量をキーボードで変更
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void txtVolume_KeyPress(object sender, KeyPressEventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            e.Handled = true;

            if (e.KeyChar == '+') {
                buf[0] = WINCTRL_CMD_VOLUP;
                streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                CurVolume += VOLUME_STEP;
                if (CurVolume > MAX_VOLUME) CurVolume = MAX_VOLUME;
            } else if (e.KeyChar == '-') {
                buf[0] = WINCTRL_CMD_VOLDOWN;
                streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                CurVolume -= VOLUME_STEP;
                if (CurVolume < MIN_VOLUME) CurVolume = MIN_VOLUME;
            } else {
                return;
            }
            txtVolume.Text = CurVolume.ToString();
            txtVolume.BackColor = COLOR_VOLUME_UNMATCH;
            TxtVolumeFocusedTime = PassedTime;
        }

        /// <summary>
        /// コントロールの状態設定：HOSTに未接続
        /// </summary>
        private void SetControlsNoConnectToHost()
        {
            btnPlay.Enabled = false;
            btnPrev.Enabled = false;
            btnNext.Enabled = false;
            btnStop.Enabled = false;
            btnSkip.Enabled = false;
            chkRandom.Enabled = false;
            txtVolume.Enabled = false;
            btnMkfl.Enabled = false;
            btnCopyfl.Enabled = false;  

            treeViewDir.Enabled = false;
        }

        /// <summary>
        /// コントロールの状態設定：HOSTに接続後
        /// </summary>
        private void SetControlsConnectToHost()
        {
            chkRandom.Enabled = true;
            treeViewDir.Enabled = true;
            btnMkfl.Enabled = true;
            btnCopyfl.Enabled = true;

            btnStartStop.Text = "停止";
        }

        /// <summary>
        /// コントロールの状態設定：再生停止中。ホストはコントローラからのコマンド待ち。
        /// </summary>
        private void SetControlsCmdWaiting()
        {
            btnPlay.Enabled = true;
            btnPrev.Enabled = false;
            btnNext.Enabled = false;
            btnStop.Enabled = false;
            btnSkip.Enabled = false;
            chkRandom.Enabled = true;
            txtVolume.Enabled = false;
            btnMkfl.Enabled = true;
            btnCopyfl.Enabled = true;
        }

        /// <summary>
        /// コントロールの状態設定：再生中
        /// </summary>
        private void SetControlsNowPlaying()
        {
            btnPlay.Enabled = false;
            btnStop.Enabled = true;
            btnSkip.Enabled = chkRandom.Checked;
            btnPrev.Enabled = true;
            btnNext.Enabled = true;
            txtVolume.Enabled = true;
            btnMkfl.Enabled = false;
            btnCopyfl.Enabled = false;  
        }

        /// <summary>
        /// コントロールの状態設定：データファイル作成中
        /// </summary>
        private void SetControlsMkfl()
        {
            btnPlay.Enabled = false;
            btnPrev.Enabled = false;
            btnNext.Enabled = false;
            btnStop.Enabled = false;
            btnSkip.Enabled = false;
            chkRandom.Enabled = false;
            txtVolume.Enabled = false;
            btnMkfl.Enabled = false;
            btnCopyfl.Enabled = false;
        }

        private void Form1_Resize(object sender, EventArgs e)
        {
            Control control = (Control)sender;

            //System.Console.WriteLine("w=" + control.Width + ",h=" + control.Height 
            //    + ",CaptionHeight=" + SystemInformation.CaptionHeight
            //    + ",BorderSize.Height=" + SystemInformation.BorderSize.Height);

            //treeViewDir.Height = control.Height - SystemInformation.CaptionHeight - 2 * SystemInformation.BorderSize.Height;
        }

        /// <summary>
        /// データファイル作成：mp3_14_mkfl実行
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnMkfl_Click(object sender, EventArgs e)
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            try {
                String dt = DateTime.Now.ToString("yyyyMMddHHmmss");
                byte [] b = Encoding.UTF8.GetBytes(dt);
                buf[0] = WINCTRL_CMD_EXECMKFL;
                for (int i=0; i<14 && i < b.Length; i++) {
                    buf[i + 1] = b[i];
                }
                streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);
                SetControlsMkfl();
            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }

        }

        /// <summary>
        /// ホストのデータファイルとローカルのデータファイルとのタイムスタンプ比較開始。
        /// タイムスタンプはDIR構成ファイルのヘッダレコードに記録されている
        /// </summary>
        /// <returns></returns>
        private void StartCompareDataTimestamp()
        {
            Byte[] buf = new byte[WINCTRL_CMDLENGTH];

            // ローカルにあるデータファイルのタイムスタンプ
            FileStream reader = null;
            try {
                reader = new FileStream(DIRSTRUCTUREFILE14, FileMode.Open, FileAccess.Read);
                byte[] RecDirList = new byte[RECLEN_DIRLIST];
                if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) == RECLEN_DIRLIST) {
                    DataTimeStampLocal = Encoding.UTF8.GetString(RecDirList, 4, 14);
                }

            } catch (Exception ex) {
            } finally {
                if (reader != null) {
                    reader.Close();
                }
            }
            System.Console.WriteLine("DataTimeStampLocal=" + DataTimeStampLocal);

            // ホストのデータファイルのタイムスタンプ取得開始
            try {
                DataTimestampMatched = false;
                WaitCmdDataTimestampFlag = true;

                buf[0] = WINCTRL_CMD_DATATIMESTAMP;
                streamToHost.Write(buf, 0, WINCTRL_CMDLENGTH);

            } catch (Exception err) {
                System.Console.WriteLine(err.Message);
            }

        }

        /// <summary>
        /// データファイルコピー
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnCopyfl_Click(object sender, EventArgs e)
        {
            try {
                SndListReader.Close();
            } catch (Exception err) {
            }
            try {
                System.Diagnostics.Process p = System.Diagnostics.Process.Start(GET_DATAFILE_BAT);
                p.WaitForExit();
            } catch (Exception err) {
                MessageBox.Show("データファイルの作成でエラー：" + err.Message);
            }
            MessageBox.Show("本プログラムを再起動してください。");
        }
    }
}
